import pytest
from unittest.mock import MagicMock

from echo import (CallbackProperty, add_callback,
                  remove_callback, delay_callback,
                  ignore_callback, callback_property,
                  HasCallbackProperties, keep_in_sync)


class Stub(object):
    prop1 = CallbackProperty()
    prop2 = CallbackProperty(5)
    prop3 = 5


class DecoratorStub(object):

    def __init__(self):
        self._val = 1

    @callback_property
    def prop(self):
        return self._val * 2

    @prop.setter
    def prop(self, value):
        self._val = value


def test_attribute_like_access():
    stub = Stub()
    assert stub.prop1 is None
    assert stub.prop2 == 5


def test_attribute_like_set():
    stub = Stub()
    stub.prop1 = 10
    assert stub.prop1 == 10


def test_class_access():
    stub = Stub()
    assert isinstance(type(stub).prop1, CallbackProperty)


def test_callback_fire_on_change():
    stub = Stub()
    test = MagicMock()
    add_callback(stub, 'prop1', test)
    stub.prop1 = 5
    test.assert_called_once_with(5)


def test_callbacks_only_called_on_value_change():
    stub = Stub()
    test = MagicMock()
    add_callback(stub, 'prop1', test)
    stub.prop1 = 5
    test.assert_called_once_with(5)
    stub.prop1 = 5
    assert test.call_count == 1


def test_callbacks_are_instance_specific():
    s1, s2 = Stub(), Stub()
    test = MagicMock()
    add_callback(s2, 'prop1', test)
    s1.prop1 = 100
    assert test.call_count == 0


def test_remove_callback():
    stub = Stub()
    test = MagicMock()
    add_callback(stub, 'prop1', test)
    remove_callback(stub, 'prop1', test)
    stub.prop1 = 5
    assert test.call_count == 0


def test_add_callback_attribute_error_on_bad_name():
    stub = Stub()
    with pytest.raises(AttributeError):
        add_callback(stub, 'bad_property', None)


def test_add_callback_type_error_if_not_calllback():
    stub = Stub()
    with pytest.raises(TypeError) as exc:
        add_callback(stub, 'prop3', None)
    assert exc.value.args[0] == "prop3 is not a CallbackProperty"


def test_remove_callback_attribute_error_on_bad_name():
    stub = Stub()
    with pytest.raises(AttributeError):
        remove_callback(stub, 'bad_property', None)


def test_remove_callback_wrong_function():
    stub = Stub()
    test = MagicMock()
    test2 = MagicMock()
    add_callback(stub, 'prop1', test)
    with pytest.raises(ValueError) as exc:
        remove_callback(stub, 'prop1', test2)
    assert exc.value.args[0].startswith('Callback function not found')


def test_remove_non_callback_property():
    stub = Stub()
    with pytest.raises(TypeError) as exc:
        remove_callback(stub, 'prop3', None)
    assert exc.value.args[0] == 'prop3 is not a CallbackProperty'


def test_remove_callback_not_found():
    stub = Stub()
    with pytest.raises(ValueError) as exc:
        remove_callback(stub, 'prop1', None)
    assert exc.value.args[0] == "Callback function not found: None"


def test_disable_callback():
    stub = Stub()
    test = MagicMock()

    add_callback(stub, 'prop1', test)
    Stub.prop1.disable(stub)

    stub.prop1 = 100
    assert test.call_count == 0

    Stub.prop1.enable(stub)
    stub.prop1 = 100
    assert test.call_count == 0  # not changed
    stub.prop1 = 200
    assert test.call_count == 1


def test_delay_callback():
    test = MagicMock()
    stub = Stub()

    add_callback(stub, 'prop1', test)
    with delay_callback(stub, 'prop1'):
        stub.prop1 = 100
        stub.prop1 = 200
        stub.prop1 = 300
        assert test.call_count == 0
    test.assert_called_once_with(300)


def test_delay_callback_nested():
    test = MagicMock()
    stub = Stub()

    add_callback(stub, 'prop1', test)
    with delay_callback(stub, 'prop1'):
        with delay_callback(stub, 'prop1'):
            stub.prop1 = 100
            stub.prop1 = 200
            stub.prop1 = 300
            assert test.call_count == 0
        assert test.call_count == 0
    test.assert_called_once_with(300)


def test_delay_callback_not_called_if_unmodified():
    test = MagicMock()
    stub = Stub()
    add_callback(stub, 'prop1', test)
    with delay_callback(stub, 'prop1'):
        pass
    assert test.call_count == 0


def test_callback_with_two_arguments():
    stub = Stub()
    stub.prop1 = 5
    on_change = MagicMock()
    add_callback(stub, 'prop1', on_change, echo_old=True)
    stub.prop1 = 10

    on_change.assert_called_once_with(5, 10)


@pytest.mark.parametrize('context_func', (delay_callback, ignore_callback))
def test_context_on_non_callback(context_func):
    stub = Stub()
    with pytest.raises(TypeError) as exc:
        with context_func(stub, 'prop3'):
            pass
    assert exc.value.args[0] == "prop3 is not a CallbackProperty"


def test_delay_multiple():
    stub = Stub()
    test = MagicMock()
    test2 = MagicMock()

    add_callback(stub, 'prop1', test)
    add_callback(stub, 'prop2', test2)

    with delay_callback(stub, 'prop1', 'prop2'):
        stub.prop1 = 50
        stub.prop1 = 100
        stub.prop2 = 200
        assert test.call_count == 0
        assert test2.call_count == 0

    test.assert_called_once_with(100)
    test2.assert_called_once_with(200)


def test_ignore_multiple():
    stub = Stub()
    test = MagicMock()
    test2 = MagicMock()

    add_callback(stub, 'prop1', test)
    add_callback(stub, 'prop2', test2)

    with ignore_callback(stub, 'prop1', 'prop2'):
        stub.prop1 = 100
        stub.prop2 = 200
        assert test.call_count == 0
        assert test2.call_count == 0

    assert test.call_count == 0
    assert test2.call_count == 0


def test_delay_only_calls_if_changed():
    stub = Stub()
    test = MagicMock()

    add_callback(stub, 'prop1', test)

    with delay_callback(stub, 'prop1'):
        pass
    assert test.call_count == 0

    val = stub.prop1
    with delay_callback(stub, 'prop1'):
        stub.prop1 = val
    assert test.call_count == 0


def test_decorator_form():
    stub = DecoratorStub()
    test = MagicMock()
    add_callback(stub, 'prop', test)

    assert stub.prop == 2

    stub.prop = 5
    test.assert_called_once_with(10)

    assert stub.prop == 10


def test_docstring():

    class Simple(object):
        a = CallbackProperty(docstring='important')

    s = Simple()
    assert type(s).a.__doc__ == 'important'


class State(HasCallbackProperties):

    a = CallbackProperty()
    b = CallbackProperty()
    c = CallbackProperty()
    d = 1


def test_class_add_remove_callback():

    state = State()

    class mockclass(object):

        def __init__(self):
            self.call_count = 0
            self.args = ()
            self.kwargs = {}

        def __call__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs
            self.call_count += 1

    test1 = mockclass()
    state.add_callback('a', test1)

    # Deliberaty adding to c twice to make sure it works fine with two callbacks
    test2 = mockclass()
    state.add_callback('c', test2)

    test3 = mockclass()
    state.add_callback('c', test3, echo_old=True)

    test4 = mockclass()
    state.add_global_callback(test4)

    state.a = 1
    assert test1.call_count == 1
    assert test1.args == (1,)
    assert test2.call_count == 0
    assert test3.call_count == 0
    assert test4.call_count == 1
    assert test4.kwargs == dict(a=1)

    state.b = 1
    assert test1.call_count == 1
    assert test2.call_count == 0
    assert test3.call_count == 0
    assert test4.call_count == 2
    assert test4.kwargs == dict(b=1)

    state.c = 1
    assert test1.call_count == 1
    assert test2.call_count == 1
    assert test4.kwargs == dict(c=1)
    assert test3.call_count == 1
    assert test3.args == (None, 1)
    assert test4.call_count == 3
    assert test4.kwargs == dict(c=1)

    state.remove_callback('a', test1)

    state.a = 2
    state.b = 2
    state.c = 2
    assert test1.call_count == 1
    assert test2.call_count == 2
    assert test3.call_count == 2
    assert test4.call_count == 6

    state.remove_callback('c', test2)

    state.a = 3
    state.b = 3
    state.c = 3
    assert test1.call_count == 1
    assert test2.call_count == 2
    assert test3.call_count == 3
    assert test4.call_count == 9

    state.remove_callback('c', test3)

    state.a = 4
    state.b = 4
    state.c = 4
    assert test1.call_count == 1
    assert test2.call_count == 2
    assert test3.call_count == 3
    assert test4.call_count == 12

    state.remove_global_callback(test4)

    state.a = 6
    state.b = 6
    state.c = 6
    assert test1.call_count == 1
    assert test2.call_count == 2
    assert test3.call_count == 3
    assert test4.call_count == 12


def test_class_is_callback_property():

    state = State()
    assert state.is_callback_property('a')
    assert state.is_callback_property('b')
    assert state.is_callback_property('c')
    assert not state.is_callback_property('d')


def test_class_add_remove_callback_invalid():

    def callback():
        pass

    state = State()
    state.z = 'banana'
    with pytest.raises(TypeError) as exc:
        state.add_callback('banana', callback)
    assert exc.value.args[0] == "attribute 'banana' is not a callback property"
    with pytest.raises(TypeError) as exc:
        state.remove_callback('banana', callback)
    assert exc.value.args[0] == "attribute 'banana' is not a callback property"


def test_keep_in_sync():

    class State1(object):
        a = CallbackProperty()
        b = CallbackProperty()

    class State2(object):
        c = CallbackProperty()

    state1 = State1()
    state2 = State2()

    state1_control = State1()
    state2_control = State2()

    s1 = keep_in_sync(state1, 'a', state1, 'b')
    s2 = keep_in_sync(state1, 'a', state2, 'c')

    state1.a = 1
    assert state1.b == 1
    assert state1_control.b is None
    assert state2.c == 1
    assert state2_control.c is None

    state1.b = 3
    assert state1.a == 3
    assert state1_control.a is None
    assert state2.c == 3
    assert state2_control.c is None

    state2.c = 5
    assert state1.a == 5
    assert state1_control.a is None
    assert state1.b == 5
    assert state1_control.b is None

    s1.disable_syncing()

    state1.a = 7
    assert state1.b == 5
    assert state1_control.b is None
    assert state2.c == 7
    assert state2_control.c is None

    s2.disable_syncing()


def test_cleanup_when_objects_destroyed():

    state = State()

    class BasicClass():

        def __init__(self, s):
            self.s = s
            self.s.add_callback('a', self.callback)
            self.raise_error = False

        def callback(self, arg):
            if self.raise_error:
                raise ValueError('Should never get here')

    def isolated(state):
        c = BasicClass(state)
        state.a = 1
        c.raise_error = True

    isolated(state)

    state.a = 2


def test_cleanup_when_objects_destroyed_kwargs():

    state = State()

    class BasicClass():

        def __init__(self, s):
            self.s = s
            self.s.add_global_callback(self.callback)
            self.raise_error = False

        def callback(self, **kwargs):
            if self.raise_error:
                raise ValueError('Should never get here')

    def isolated(state):
        c = BasicClass(state)
        state.a = 1
        c.raise_error = True

    isolated(state)

    state.a = 2


def test_delay_global_callback():

    # Regression test to make sure that delay_callback works for global
    # callbacks too.

    state = State()

    test1 = MagicMock()
    state.add_callback('a', test1)

    test2 = MagicMock()
    state.add_global_callback(test2)

    with delay_callback(state, 'a'):
        state.a = 100
        assert test1.call_count == 0
        assert test2.call_count == 0

    test1.assert_called_once_with(100)
    test2.assert_called_once_with(a=100)

    test2.reset_mock()

    with delay_callback(state, 'a'):
        state.b = 200
        assert test2.call_count == 1

    test2.assert_called_once_with(b=200)

    test2.reset_mock()

    with delay_callback(state, 'a', 'b'):
        state.a = 300
        state.b = 400
        assert test2.call_count == 0

    test2.assert_called_once_with(a=300, b=400)


def test_delay_global_callback_stub():

    # Make sure that adding the global callback delay functionality doesn't
    # break things when we are dealing with a plain class without HasCallbackProperties

    stub = Stub()

    test1 = MagicMock()
    add_callback(stub, 'prop1', test1)

    with delay_callback(stub, 'prop1'):
        stub.prop1 = 100
        assert test1.call_count == 0

    test1.assert_called_once_with(100)


def test_ignore_global_callback():

    # Regression test to make sure that ignore_callback works for global
    # callbacks too.

    state = State()

    test1 = MagicMock()
    state.add_callback('a', test1)

    test2 = MagicMock()
    state.add_global_callback(test2)

    with ignore_callback(state, 'a'):
        state.a = 100
        assert test1.call_count == 0
        assert test2.call_count == 0

    assert test1.call_count == 0
    assert test2.call_count == 0

    test2.reset_mock()

    with ignore_callback(state, 'a'):
        state.b = 200
        assert test2.call_count == 1

    test2.assert_called_once_with(b=200)

    test2.reset_mock()

    with ignore_callback(state, 'a', 'b'):
        state.a = 300
        state.b = 400
        assert test2.call_count == 0

    assert test2.call_count == 0


def test_ignore_global_callback_stub():

    # Make sure that adding the global callback ignore functionality doesn't
    # break things when we are dealing with a plain class without HasCallbackProperties

    stub = Stub()

    test1 = MagicMock()
    add_callback(stub, 'prop1', test1)

    with ignore_callback(stub, 'prop1'):
        stub.prop1 = 100
        assert test1.call_count == 0

    assert test1.call_count == 0


def test_delay_in_delayed_callback():

    # Regression test for a bug that occurred if a delayed callback included
    # a delay itself.

    state = State()

    def callback(*args, **kwargs):
        with delay_callback(state, 'a'):
            state.a = 2

    state.add_callback('a', callback)

    with delay_callback(state, 'a', 'b'):
        state.a = 100


def test_ignore_in_ignored_callback():

    # Regression test for a bug that occurred if a delayed callback included
    # a delay itself.

    state = State()

    def callback(*args, **kwargs):
        with ignore_callback(state, 'a'):
            state.a = 2

    state.add_callback('a', callback)

    with ignore_callback(state, 'a', 'b'):
        state.a = 100



from __future__ import absolute_import, division, print_function

try:
    import qtpy  # noqa
except Exception:
    QT_INSTALLED = False
else:
    QT_INSTALLED = True

qapp = None


def get_qapp():
    global qapp
    from qtpy import QtWidgets
    qapp = QtWidgets.QApplication.instance()
    if qapp is None:
        qapp = QtWidgets.QApplication([''])
    return qapp


def pytest_configure(config):
    if QT_INSTALLED:
        from qtpy import QtWidgets  # noqa
        app = get_qapp()  # noqa


def pytest_unconfigure(config):
    if QT_INSTALLED:
        global qapp
        qapp.exit()
        qapp = None



import pytest
from unittest.mock import MagicMock

from echo import (CallbackProperty,
                  ListCallbackProperty,
                  DictCallbackProperty,
                  HasCallbackProperties)


class StubList(HasCallbackProperties):
    prop1 = ListCallbackProperty()
    prop2 = ListCallbackProperty()


class StubDict(HasCallbackProperties):
    prop1 = DictCallbackProperty()
    prop2 = DictCallbackProperty()


class Simple(HasCallbackProperties):
    a = CallbackProperty()


def test_list_normal_callback():

    stub = StubList()

    test1 = MagicMock()
    stub.add_callback('prop1', test1)

    stub.prop1 = [3]
    assert test1.call_count == 1

    stub.prop2 = [4]
    assert test1.call_count == 1


def test_list_invalid():

    stub = StubList()
    with pytest.raises(TypeError) as exc:
        stub.prop1 = 'banana'
    assert exc.value.args[0] == "callback property should be a list"


def test_list_default_empty():

    stub = StubList()
    stub.prop1.append(1)


def test_list_explicit_default():

    class TestList(HasCallbackProperties):
        prop = ListCallbackProperty(default=[1, 2, 3])

    test1 = TestList()
    assert test1.prop == [1, 2, 3]

    test2 = TestList()
    assert test2.prop == [1, 2, 3]

    test1.prop.append(4)
    assert test1.prop == [1, 2, 3, 4]
    assert test2.prop == [1, 2, 3]


def test_list_change_callback():

    stub = StubList()

    test1 = MagicMock()
    stub.add_callback('prop1', test1)

    assert test1.call_count == 0

    stub.prop1.append(3)
    assert test1.call_count == 1
    assert stub.prop1 == [3]

    stub.prop1.clear()
    assert test1.call_count == 2
    assert stub.prop1 == []

    stub.prop1.extend([1, 2, 3])
    assert test1.call_count == 3
    assert stub.prop1 == [1, 2, 3]

    stub.prop1.insert(0, -1)
    assert test1.call_count == 4
    assert stub.prop1 == [-1, 1, 2, 3]

    p = stub.prop1.pop()
    assert test1.call_count == 5
    assert p == 3
    assert stub.prop1 == [-1, 1, 2]

    stub.prop1.remove(1)
    assert test1.call_count == 6
    assert stub.prop1 == [-1, 2]

    stub.prop1.reverse()
    assert test1.call_count == 7
    assert stub.prop1 == [2, -1]

    stub.prop1.sort()
    assert test1.call_count == 8
    assert stub.prop1 == [-1, 2]

    stub.prop1[0] = 3
    assert test1.call_count == 9
    assert stub.prop1 == [3, 2]

    stub.prop1[:] = []
    assert test1.call_count == 10
    assert stub.prop1 == []


def test_state_in_a_list():

    stub = StubList()

    state1 = Simple()
    state2 = Simple()
    state3 = Simple()
    state4 = Simple()
    state5 = Simple()

    # Add three of the state objects to the list in different ways
    stub.prop1.append(state1)
    stub.prop1.extend([state2, 3.4])
    stub.prop1.insert(1, state3)
    stub.prop1[3] = state4

    # Check that all states except state 5 have a callback added
    assert len(state1._global_callbacks) == 1
    assert len(state2._global_callbacks) == 1
    assert len(state3._global_callbacks) == 1
    assert len(state4._global_callbacks) == 1
    assert len(state5._global_callbacks) == 0

    # Add a callback to the main list
    callback = MagicMock()
    stub.add_callback('prop1', callback)

    # Check that modifying attributes of the state objects triggers the list
    # callback.
    assert callback.call_count == 0
    state1.a = 1
    assert callback.call_count == 1
    state2.a = 1
    assert callback.call_count == 2
    state3.a = 1
    assert callback.call_count == 3
    state4.a = 1
    assert callback.call_count == 4
    state5.a = 1
    assert callback.call_count == 4

    # Remove one of the state objects and try again
    stub.prop1.pop(0)
    assert callback.call_count == 5
    assert len(state1._global_callbacks) == 0
    assert len(state2._global_callbacks) == 1
    assert len(state3._global_callbacks) == 1
    assert len(state4._global_callbacks) == 1
    assert len(state5._global_callbacks) == 0

    # Now modifying state1 sholdn't affect the call cont
    state1.a = 2
    assert callback.call_count == 5
    state2.a = 2
    assert callback.call_count == 6
    state3.a = 2
    assert callback.call_count == 7
    state4.a = 2
    assert callback.call_count == 8
    state5.a = 2
    assert callback.call_count == 8

    # Remove again this time using remove
    stub.prop1.remove(state2)
    assert callback.call_count == 9
    assert len(state1._global_callbacks) == 0
    assert len(state2._global_callbacks) == 0
    assert len(state3._global_callbacks) == 1
    assert len(state4._global_callbacks) == 1
    assert len(state5._global_callbacks) == 0

    # Now modifying state2 sholdn't affect the call cont
    state1.a = 3
    assert callback.call_count == 9
    state2.a = 3
    assert callback.call_count == 9
    state3.a = 3
    assert callback.call_count == 10
    state4.a = 3
    assert callback.call_count == 11
    state5.a = 3
    assert callback.call_count == 11

    # Remove using item access
    stub.prop1[1] = 3.3
    assert callback.call_count == 12
    assert len(state1._global_callbacks) == 0
    assert len(state2._global_callbacks) == 0
    assert len(state3._global_callbacks) == 1
    assert len(state4._global_callbacks) == 0
    assert len(state5._global_callbacks) == 0

    # Now modifying state4 sholdn't affect the call cont
    state1.a = 4
    assert callback.call_count == 12
    state2.a = 4
    assert callback.call_count == 12
    state3.a = 4
    assert callback.call_count == 13
    state4.a = 4
    assert callback.call_count == 13
    state5.a = 4
    assert callback.call_count == 13

    # Now use slice access to remove state3 and add state5 in one go
    stub.prop1[0:2] = [2.2, state5]
    assert callback.call_count == 14
    assert len(state1._global_callbacks) == 0
    assert len(state2._global_callbacks) == 0
    assert len(state3._global_callbacks) == 0
    assert len(state4._global_callbacks) == 0
    assert len(state5._global_callbacks) == 1

    # Now only modifying state5 should have an effect
    state1.a = 5
    assert callback.call_count == 14
    state2.a = 5
    assert callback.call_count == 14
    state3.a = 5
    assert callback.call_count == 14
    state4.a = 5
    assert callback.call_count == 14
    state5.a = 5
    assert callback.call_count == 15

    # On Python 3, check that clear does the right thing
    stub.prop1.clear()
    assert callback.call_count == 16
    assert len(state1._global_callbacks) == 0
    assert len(state2._global_callbacks) == 0
    assert len(state3._global_callbacks) == 0
    assert len(state4._global_callbacks) == 0
    assert len(state5._global_callbacks) == 0

    # Now the callback should never be called
    state1.a = 6
    assert callback.call_count == 16
    state2.a = 6
    assert callback.call_count == 16
    state3.a = 6
    assert callback.call_count == 16
    state4.a = 6
    assert callback.call_count == 16
    state5.a = 6
    assert callback.call_count == 16


def test_list_nested_callbacks():

    stub1 = StubList()
    stub2 = StubList()
    simple = Simple()
    stub1.prop1.append(stub2)
    stub1.prop1.append(simple)

    test1 = MagicMock()

    stub1.add_callback('prop1', test1)

    stub2.prop1.append(2)
    assert test1.call_count == 1

    simple.a = 'banana!'
    assert test1.call_count == 2


def test_dict_normal_callback():

    stub = StubDict()

    test1 = MagicMock()
    stub.add_callback('prop1', test1)

    stub.prop1 = {'a': 1}
    assert test1.call_count == 1

    stub.prop2 = {'b': 2}
    assert test1.call_count == 1


def test_dict_invalid():

    stub = StubDict()
    with pytest.raises(TypeError) as exc:
        stub.prop1 = 'banana'
    assert exc.value.args[0] == "Callback property should be a dictionary."


def test_dict_default_empty():

    stub = StubDict()
    stub.prop1['a'] = 1


def test_dict_explicit_default():

    class TestDict(HasCallbackProperties):
        prop = DictCallbackProperty(default={'a': 1, 'b': 2})

    test1 = TestDict()
    assert test1.prop == {'a': 1, 'b': 2}

    test2 = TestDict()
    assert test2.prop == {'a': 1, 'b': 2}

    test1.prop['c'] = 3
    assert test1.prop == {'a': 1, 'b': 2, 'c': 3}
    assert test2.prop == {'a': 1, 'b': 2}


def test_dict_change_callback():

    stub = StubDict()

    test1 = MagicMock()
    stub.add_callback('prop1', test1)

    assert test1.call_count == 0

    stub.prop1['a'] = 1
    assert test1.call_count == 1
    assert stub.prop1 == {'a': 1}

    stub.prop1.clear()
    assert test1.call_count == 2
    assert stub.prop1 == {}

    stub.prop1.update({'b': 2, 'c': 3, 'd': 4})
    assert test1.call_count == 3
    assert stub.prop1 == {'b': 2, 'c': 3, 'd': 4}

    p = stub.prop1.pop('b')
    assert test1.call_count == 4
    assert p == 2
    assert stub.prop1 == {'c': 3, 'd': 4}

    k, v = stub.prop1.popitem()
    assert test1.call_count == 5
    assert stub.prop1 == {'c': 3} if k == 'd' else {'d': 4}
    remaining_key = 'c' if k == 'd' else 'd'

    stub.prop1[remaining_key] = 6
    assert test1.call_count == 6
    assert stub.prop1 == {remaining_key: 6}

    stub.prop1.clear()
    assert test1.call_count == 7
    assert stub.prop1 == {}


def test_state_in_a_dict():

    stub = StubDict()

    state1 = Simple()
    state2 = Simple()
    state3 = Simple()

    # Add three of the state objects to the dict in different ways
    stub.prop1['a'] = state1
    stub.prop1.update({'b': state2})

    # Check that all states except state 3 have a callback added
    assert len(state1._global_callbacks) == 1
    assert len(state2._global_callbacks) == 1
    assert len(state3._global_callbacks) == 0

    # Add a callback to the main dict
    callback = MagicMock()
    stub.add_callback('prop1', callback)

    # Check that modifying attributes of the state objects triggers the list
    # callback.
    assert callback.call_count == 0
    state1.a = 1
    assert callback.call_count == 1
    state2.a = 1
    assert callback.call_count == 2
    state3.a = 1
    assert callback.call_count == 2

    # Remove one of the state objects and try again
    stub.prop1.pop('a')
    assert callback.call_count == 3
    assert len(state1._global_callbacks) == 0
    assert len(state2._global_callbacks) == 1
    assert len(state3._global_callbacks) == 0

    # Now modifying state1 sholdn't affect the call count
    state1.a = 2
    assert callback.call_count == 3
    state2.a = 2
    assert callback.call_count == 4
    state3.a = 2
    assert callback.call_count == 4

    # Remove using item access and add state3
    stub.prop1['b'] = 3.3
    stub.prop1['c'] = state3
    assert callback.call_count == 6
    assert len(state1._global_callbacks) == 0
    assert len(state2._global_callbacks) == 0
    assert len(state3._global_callbacks) == 1

    # Now modifying state2 sholdn't affect the call cont
    state1.a = 4
    assert callback.call_count == 6
    state2.a = 4
    assert callback.call_count == 6
    state3.a = 4
    assert callback.call_count == 7

    # Check that clear does the right thing
    stub.prop1.clear()
    assert callback.call_count == 8
    assert len(state1._global_callbacks) == 0
    assert len(state2._global_callbacks) == 0
    assert len(state3._global_callbacks) == 0

    # Now the callback should never be called
    state1.a = 6
    assert callback.call_count == 8
    state2.a = 6
    assert callback.call_count == 8
    state3.a = 6
    assert callback.call_count == 8


def test_dict_nested_callbacks():

    stub1 = StubDict()
    stub2 = StubDict()
    simple = Simple()
    stub1.prop1['a'] = stub2
    stub1.prop1['b'] = simple

    test1 = MagicMock()

    stub1.add_callback('prop1', test1)

    stub2.prop1['c'] = 2
    assert test1.call_count == 1

    simple.a = 'banana!'
    assert test1.call_count == 2


def test_multiple_nested_list_and_dict():

    stubd = StubDict()

    test1 = MagicMock()
    stubd.add_callback('prop1', test1)

    stubd.prop1 = {'a': {'b': {'c': 1}}}
    assert test1.call_count == 1

    stubd.prop1['a']['b']['c'] = 2
    assert test1.call_count == 2

    stubd.prop1['a']['b']['c'] = {'d': 3}
    assert test1.call_count == 3

    stubd.prop1['a']['b']['c']['d'] = 4
    assert test1.call_count == 4

    stubd.prop1['a']['b'] = [1, 2, 3]
    assert test1.call_count == 5

    stubd.prop1['a']['b'][1] = {'e': 6}
    assert test1.call_count == 6

    stubd.prop1['a']['b'][1]['e'] = 7
    assert test1.call_count == 7

    stubd.prop1['a'] = 2
    assert test1.call_count == 8

    stubl = StubList()

    test2 = MagicMock()
    stubl.add_callback('prop1', test2)

    stubl.prop1 = [1, 2, {'a': 3, 'b': [4, 5, {'c': 6}]}]
    assert test2.call_count == 1

    stubl.prop1[2]['b'][2]['c'] = 5
    assert test2.call_count == 2

    stubl.prop1[2]['b'][2] = {'d': 7}
    assert test2.call_count == 3

    stubl.prop1[2]['b'][2]['d'] = 2
    assert test2.call_count == 4

    stubl.prop1[2]['b'][2] = [1, 2]
    assert test2.call_count == 5

    stubl.prop1[2]['b'][2][0] = 3
    assert test2.call_count == 6

    stubl.prop1[2] = [1, 2]
    assert test2.call_count == 7

    stubl.prop1[2][1] = 3
    assert test2.call_count == 8


def test_list_additional_callbacks():

    stub = StubList()

    test1 = MagicMock()
    test2 = MagicMock()
    stub.add_callback('prop1', test1)

    stub.prop1 = [3]
    stub.prop2 = [4]

    assert test1.call_count == 1
    assert test2.call_count == 0

    stub.prop1.callbacks.append(test2)

    stub.prop2.append(5)
    assert test1.call_count == 1
    assert test2.call_count == 0

    stub.prop1.append(2)
    assert test1.call_count == 2
    assert test2.call_count == 1

    stub.prop1.callbacks.remove(test2)
    stub.prop1.append(4)
    assert test1.call_count == 3
    assert test2.call_count == 1


def test_dict_additional_callbacks():

    stub = StubDict()

    test1 = MagicMock()
    test2 = MagicMock()
    stub.add_callback('prop1', test1)

    stub.prop1 = {'a': 1}
    stub.prop2 = {'b': 2}

    assert test1.call_count == 1
    assert test2.call_count == 0

    stub.prop1.callbacks.append(test2)

    stub.prop2['c'] = 3
    assert test1.call_count == 1
    assert test2.call_count == 0

    stub.prop1['d'] = 4
    assert test1.call_count == 2
    assert test2.call_count == 1

    stub.prop1.callbacks.remove(test2)
    stub.prop1['e'] = 5
    assert test1.call_count == 3
    assert test2.call_count == 1



from . import CallbackProperty, HasCallbackProperties
from .callback_container import CallbackContainer

__all__ = ['CallbackList', 'CallbackDict',
           'ListCallbackProperty', 'DictCallbackProperty']


class ContainerMixin:

    def _prepare_add(self, value):
        if isinstance(value, list):
            value = CallbackList(self.notify_all, value)
        elif isinstance(value, dict):
            value = CallbackDict(self.notify_all, value)
        if isinstance(value, HasCallbackProperties):
            value.add_global_callback(self.notify_all)
        elif isinstance(value, (CallbackList, CallbackDict)):
            value.callback = self.notify_all
        return value

    def _cleanup_remove(self, value):
        if isinstance(value, HasCallbackProperties):
            value.remove_global_callback(self.notify_all)
        elif isinstance(value, (CallbackList, CallbackDict)):
            value.callbacks.remove(self.notify_all)


class CallbackList(list, ContainerMixin):
    """
    A list that calls a callback function when it is modified.

    The first argument should be the callback function (which takes no
    arguments), and subsequent arguments are as for `list`.
    """

    def __init__(self, callback, *args, **kwargs):
        super(CallbackList, self).__init__(*args, **kwargs)
        self.callbacks = CallbackContainer()
        self.callbacks.append(callback)
        for index, value in enumerate(self):
            super().__setitem__(index, self._prepare_add(value))

    def notify_all(self, *args, **kwargs):
        for callback in self.callbacks:
            callback(*args, **kwargs)

    def __repr__(self):
        return "<CallbackList with {0} elements>".format(len(self))

    def append(self, value):
        super(CallbackList, self).append(self._prepare_add(value))
        self.notify_all()

    def extend(self, iterable):
        iterable = [self._prepare_add(value) for value in iterable]
        super(CallbackList, self).extend(iterable)
        self.notify_all()

    def insert(self, index, value):
        super(CallbackList, self).insert(index, self._prepare_add(value))
        self.notify_all()

    def pop(self, index=-1):
        result = super(CallbackList, self).pop(index)
        self._cleanup_remove(result)
        self.notify_all()
        return result

    def remove(self, value):
        super(CallbackList, self).remove(value)
        self._cleanup_remove(value)
        self.notify_all()

    def reverse(self):
        super(CallbackList, self).reverse()
        self.notify_all()

    def sort(self, key=None, reverse=False):
        super(CallbackList, self).sort(key=key, reverse=reverse)
        self.notify_all()

    def __setitem__(self, slc, new_value):

        old_values = self[slc]
        if not isinstance(slc, slice):
            old_values = [old_values]

        for old_value in old_values:
            self._cleanup_remove(old_value)

        if isinstance(slc, slice):
            new_value = [self._prepare_add(value) for value in new_value]
        else:
            new_value = self._prepare_add(new_value)

        super(CallbackList, self).__setitem__(slc, new_value)
        self.notify_all()

    def clear(self):
        for item in self:
            self._cleanup_remove(item)
        super(CallbackList, self).clear()
        self.notify_all()


class CallbackDict(dict, ContainerMixin):
    """
    A dictionary that calls a callback function when it is modified.

    The first argument should be the callback function (which takes no
    arguments), and subsequent arguments are passed to `dict`.
    """

    def __init__(self, callback, *args, **kwargs):
        super(CallbackDict, self).__init__(*args, **kwargs)
        self.callbacks = CallbackContainer()
        self.callbacks.append(callback)
        for key, value in self.items():
            super().__setitem__(key, self._prepare_add(value))

    def notify_all(self, *args, **kwargs):
        for callback in self.callbacks:
            callback(*args, **kwargs)

    def clear(self):
        for value in self.values():
            self._cleanup_remove(value)
        super().clear()
        self.notify_all()

    def popitem(self):
        result = super().popitem()
        self._cleanup_remove(result)
        self.notify_all()
        return result

    def update(self, *args, **kwargs):
        values = {}
        values.update(*args, **kwargs)
        for key, value in values.items():
            values[key] = self._prepare_add(value)
        super().update(values)
        self.notify_all()

    def pop(self, *args, **kwargs):
        result = super().pop(*args, **kwargs)
        self._cleanup_remove(result)
        self.notify_all()
        return result

    def __setitem__(self, key, value):
        if key in self:
            self._cleanup_remove(self[key])
        super().__setitem__(key, self._prepare_add(value))
        self.notify_all()

    def __repr__(self):
        return f"<CallbackDict with {len(self)} elements>"


class dynamic_callback:

    function = None

    def __call__(self, *args, **kwargs):
        self.function(*args, **kwargs)


class ListCallbackProperty(CallbackProperty):
    """
    A list property that calls callbacks when its contents are modified
    """

    def _default_getter(self, instance, owner=None):
        if instance not in self._values:
            self._default_setter(instance, self._default or [])
        return super(ListCallbackProperty, self)._default_getter(instance, owner)

    def _default_setter(self, instance, value):

        if not isinstance(value, list):
            raise TypeError('callback property should be a list')

        dcb = dynamic_callback()

        wrapped_list = CallbackList(dcb, value)

        def callback(*args, **kwargs):
            self.notify(instance, wrapped_list, wrapped_list)

        dcb.function = callback

        super(ListCallbackProperty, self)._default_setter(instance, wrapped_list)


class DictCallbackProperty(CallbackProperty):
    """
    A dictionary property that calls callbacks when its contents are modified
    """
    def _default_getter(self, instance, owner=None):
        if instance not in self._values:
            self._default_setter(instance, self._default or {})
        return super()._default_getter(instance, owner)

    def _default_setter(self, instance, value):

        if not isinstance(value, dict):
            raise TypeError("Callback property should be a dictionary.")

        dcb = dynamic_callback()

        wrapped_dict = CallbackDict(dcb, value)

        def callback(*args, **kwargs):
            self.notify(instance, wrapped_dict, wrapped_dict)

        dcb.function = callback

        super()._default_setter(instance, wrapped_dict)



import pytest
from unittest.mock import MagicMock

from qtpy import QtWidgets

from echo import CallbackProperty
from echo.qt.connect import (connect_checkable_button, connect_text,
                             connect_combo_data, connect_combo_text,
                             connect_float_text, connect_value, connect_button,
                             UserDataWrapper)


def test_connect_checkable_button():

    class Test(object):
        a = CallbackProperty()
        b = CallbackProperty(True)

    t = Test()

    box1 = QtWidgets.QCheckBox()
    c1 = connect_checkable_button(t, 'a', box1)  # noqa

    box1.setChecked(True)
    assert t.a

    box1.setChecked(False)
    assert not t.a

    t.a = True
    assert box1.isChecked()

    t.a = False
    assert not box1.isChecked()

    # Make sure that the default value of the callback property is recognized

    box2 = QtWidgets.QCheckBox()
    connect_checkable_button(t, 'b', box2)

    assert box2.isChecked()


def test_connect_text():

    class Test(object):
        a = CallbackProperty()
        b = CallbackProperty()

    t = Test()

    box = QtWidgets.QLineEdit()
    c1 = connect_text(t, 'a', box)  # noqa

    label = QtWidgets.QLabel()
    c2 = connect_text(t, 'b', label)  # noqa

    box.setText('test1')
    box.editingFinished.emit()
    assert t.a == 'test1'

    t.a = 'test3'
    assert box.text() == 'test3'

    t.b = 'test4'
    assert label.text() == 'test4'


def test_connect_combo():

    class Test(object):
        a = CallbackProperty()
        b = CallbackProperty()

    t = Test()

    combo = QtWidgets.QComboBox()
    combo.addItem('label1', UserDataWrapper(4))
    combo.addItem('label2', UserDataWrapper(3.5))

    c1 = connect_combo_text(t, 'a', combo)  # noqa
    c2 = connect_combo_data(t, 'b', combo)  # noqa

    combo.setCurrentIndex(1)
    assert t.a == 'label2'
    assert t.b == 3.5

    combo.setCurrentIndex(0)
    assert t.a == 'label1'
    assert t.b == 4

    combo.setCurrentIndex(-1)
    assert t.a is None
    assert t.b is None

    t.a = 'label2'
    assert combo.currentIndex() == 1

    t.a = 'label1'
    assert combo.currentIndex() == 0

    with pytest.raises(ValueError) as exc:
        t.a = 'label3'
    assert exc.value.args[0] == 'label3 not found in combo box'

    t.a = None
    assert combo.currentIndex() == -1

    t.b = 3.5
    assert combo.currentIndex() == 1

    t.b = 4
    assert combo.currentIndex() == 0

    with pytest.raises(ValueError) as exc:
        t.b = 2
    assert exc.value.args[0] == '2 not found in combo box'

    t.b = None
    assert combo.currentIndex() == -1


def test_connect_float_text():

    class Test(object):
        a = CallbackProperty()
        b = CallbackProperty()
        c = CallbackProperty()

    t = Test()

    line1 = QtWidgets.QLineEdit()
    line2 = QtWidgets.QLineEdit()
    line3 = QtWidgets.QLabel()

    def fmt_func(x):
        return str(int(round(x)))

    c1 = connect_float_text(t, 'a', line1)  # noqa
    c2 = connect_float_text(t, 'b', line2, fmt="{:5.2f}")  # noqa
    c3 = connect_float_text(t, 'c', line3, fmt=fmt_func)  # noqa

    for line in (line1, line2):

        line1.setText('1.0')
        line1.editingFinished.emit()
        assert t.a == 1.0

        line1.setText('banana')
        line1.editingFinished.emit()
        assert t.a == 0.0

    t.a = 3.
    assert line1.text() == '3'

    t.b = 5.211
    assert line2.text() == ' 5.21'

    t.c = -2.222
    assert line3.text() == '-2'


def test_connect_value():

    class Test(object):
        a = CallbackProperty()
        b = CallbackProperty()
        c = CallbackProperty()

    t = Test()

    slider = QtWidgets.QSlider()
    slider.setMinimum(0)
    slider.setMaximum(100)

    c1 = connect_value(t, 'a', slider)  # noqa
    c2 = connect_value(t, 'b', slider, value_range=(0, 10))  # noqa

    with pytest.raises(Exception) as exc:
        connect_value(t, 'c', slider, log=True)
    assert exc.value.args[0] == "log option can only be set if value_range is given"

    c3 = connect_value(t, 'c', slider, value_range=(0.01, 100), log=True)  # noqa

    slider.setValue(25)
    assert t.a == 25
    assert t.b == 2.5
    assert t.c == 0.1

    t.a = 30
    assert slider.value() == 30

    t.b = 8.5
    assert slider.value() == 85

    t.c = 10
    assert slider.value() == 75


def test_connect_button():

    class Example(object):
        a = MagicMock()

    e = Example()

    button = QtWidgets.QPushButton('OK')

    connect_button(e, 'a', button)

    assert e.a.call_count == 0
    button.clicked.emit()
    assert e.a.call_count == 1



import pytest
from unittest.mock import MagicMock

from ..core import HasCallbackProperties, delay_callback
from ..selection import SelectionCallbackProperty, ChoiceSeparator


class Example(HasCallbackProperties):
    a = SelectionCallbackProperty()
    b = SelectionCallbackProperty(default_index=1)
    c = SelectionCallbackProperty(default_index=-1)


class TestSelectionCallbackProperty():

    def setup_method(self, method):
        self.state = Example()
        self.a = Example.a
        self.b = Example.b
        self.c = Example.c

    def test_set_choices(self):
        # make sure that default_index is respected
        self.a.set_choices(self.state, [1, 2])
        self.b.set_choices(self.state, [1, 2])
        self.c.set_choices(self.state, [1, 2, 3])
        assert self.state.a == 1
        assert self.state.b == 2
        assert self.state.c == 3
        self.a.set_choices(self.state, None)
        assert self.state.a is None

    def test_set_value(self):

        # Since there are no choices set yet, the properties cannot be set to
        # any values yet.
        with pytest.raises(ValueError) as exc:
            self.state.a = 1
        assert exc.value.args[0] == "value 1 is not in valid choices: []"

        self.a.set_choices(self.state, [1, 2])

        # The following should work since it is a valid choice
        self.state.a = 2

        # Again if we try and set to an invalid value, things break
        with pytest.raises(ValueError) as exc:
            self.state.a = 3
        assert exc.value.args[0] == "value 3 is not in valid choices: [1, 2]"

    def test_value_constant(self):

        # Make sure that the value is preserved if possible

        self.a.set_choices(self.state, [1, 2])
        self.state.a = 2

        self.a.set_choices(self.state, [2, 5, 3])
        assert self.state.a == 2

        self.a.set_choices(self.state, [1, 4, 5])
        assert self.state.a == 1

        self.a.set_choices(self.state, [4, 2, 1])
        assert self.state.a == 1

    def test_get_choices(self):

        self.a.set_choices(self.state, [1, 2])
        self.a.get_choices == [1, 2]

        self.a.set_choices(self.state, [2, 5, 3])
        self.a.get_choices == [2, 5, 3]

    def test_display_func(self):

        separator = ChoiceSeparator('header')

        self.a.set_choices(self.state, [separator, 1, 2])
        self.a.get_choice_labels(self.state) == ['header', '1', '2']
        assert self.a.get_display_func(self.state) is None
        assert self.b.get_display_func(self.state) is None

        def val(x):
            return 'val{0}'.format(x)

        self.a.set_display_func(self.state, val)
        self.a.get_choice_labels(self.state) == ['header', 'val1', 'val2']
        assert self.a.get_display_func(self.state) is val
        assert self.b.get_display_func(self.state) is None

    def test_callbacks(self):

        # Make sure that callbacks are called when either choices or selection
        # are changed

        func = MagicMock()
        self.state.add_callback('a', func)

        self.a.set_choices(self.state, [1, 2, 3])
        assert func.called_once_with(1)

        self.state.a = 2
        assert func.called_once_with(2)

        self.a.set_choices(self.state, [4, 5, 6])
        assert func.called_once_with(4)

    def test_choice_separator(self):

        separator = ChoiceSeparator('header')
        self.a.set_choices(self.state, [separator, 1, 2])

        assert self.state.a == 1

        separator = ChoiceSeparator('header')
        self.a.set_choices(self.state, [separator])

        assert self.state.a is None

    def test_delay(self):

        func = MagicMock()
        self.state.add_callback('a', func)

        # Here we set the choices and as a result the selection changes from
        # None to a value, so the callback is called after the delay block
        with delay_callback(self.state, 'a'):
            self.a.set_choices(self.state, [4, 5, 6])
            assert func.call_count == 0
        assert func.called_once_with(4)
        func.reset_mock()

        # Check that the callback gets called even if only the choices
        # but not the selection are changed in a delay block
        with delay_callback(self.state, 'a'):
            self.a.set_choices(self.state, [1, 2, 4])
            assert func.call_count == 0
        assert func.called_once_with(4)
        func.reset_mock()



import pytest
import numpy as np

from qtpy import QtWidgets

from echo.core import CallbackProperty
from echo.selection import SelectionCallbackProperty, ChoiceSeparator
from echo.qt.connect import connect_combo_selection


class Example(object):
    a = SelectionCallbackProperty(default_index=1)
    b = CallbackProperty()


def test_connect_combo_selection():

    t = Example()

    a_prop = getattr(type(t), 'a')
    a_prop.set_choices(t, [4, 3.5])
    a_prop.set_display_func(t, lambda x: 'value: {0}'.format(x))

    combo = QtWidgets.QComboBox()

    c1 = connect_combo_selection(t, 'a', combo)  # noqa

    assert combo.itemText(0) == 'value: 4'
    assert combo.itemText(1) == 'value: 3.5'
    assert combo.itemData(0).data == 4
    assert combo.itemData(1).data == 3.5

    combo.setCurrentIndex(1)
    assert t.a == 3.5

    combo.setCurrentIndex(0)
    assert t.a == 4

    combo.setCurrentIndex(-1)
    assert t.a is None

    t.a = 3.5
    assert combo.currentIndex() == 1

    t.a = 4
    assert combo.currentIndex() == 0

    with pytest.raises(ValueError) as exc:
        t.a = 2
    assert exc.value.args[0] == 'value 2 is not in valid choices: [4, 3.5]'

    t.a = None
    assert combo.currentIndex() == -1

    # Changing choices should change Qt combo box. Let's first try with a case
    # in which there is a matching data value in the new combo box

    t.a = 3.5
    assert combo.currentIndex() == 1

    a_prop.set_choices(t, (4, 5, 3.5))
    assert combo.count() == 3

    assert t.a == 3.5
    assert combo.currentIndex() == 2

    assert combo.itemText(0) == 'value: 4'
    assert combo.itemText(1) == 'value: 5'
    assert combo.itemText(2) == 'value: 3.5'
    assert combo.itemData(0).data == 4
    assert combo.itemData(1).data == 5
    assert combo.itemData(2).data == 3.5

    # Now we change the choices so that there is no matching data - in this case
    # the index should change to that given by default_index

    a_prop.set_choices(t, (4, 5, 6))

    assert t.a == 5
    assert combo.currentIndex() == 1
    assert combo.count() == 3

    assert combo.itemText(0) == 'value: 4'
    assert combo.itemText(1) == 'value: 5'
    assert combo.itemText(2) == 'value: 6'
    assert combo.itemData(0).data == 4
    assert combo.itemData(1).data == 5
    assert combo.itemData(2).data == 6

    # Finally, if there are too few choices for the default_index to be valid,
    # pick the last item in the combo

    a_prop.set_choices(t, (9,))

    assert t.a == 9
    assert combo.currentIndex() == 0
    assert combo.count() == 1

    assert combo.itemText(0) == 'value: 9'
    assert combo.itemData(0).data == 9

    # Now just make sure that ChoiceSeparator works

    separator = ChoiceSeparator('header')
    a_prop.set_choices(t, (separator, 1, 2))

    assert combo.count() == 3
    assert combo.itemText(0) == 'header'
    assert combo.itemData(0).data is separator

    # And setting choices to an empty iterable shouldn't cause issues

    a_prop.set_choices(t, ())
    assert combo.count() == 0

    # Try including an array in the choices
    a_prop.set_choices(t, (4, 5, np.array([1, 2, 3])))


def test_connect_combo_selection_invalid():

    t = Example()

    combo = QtWidgets.QComboBox()

    with pytest.raises(TypeError) as exc:
        connect_combo_selection(t, 'b', combo)
    assert exc.value.args[0] == 'connect_combo_selection requires a SelectionCallbackProperty'



from qtpy import QtWidgets, QtGui

from echo.qt.autoconnect import autoconnect_callbacks_to_qt
from echo import CallbackProperty
from echo.qt.connect import UserDataWrapper


def test_autoconnect_callbacks_to_qt():

    class Data(object):
        pass

    data1 = Data()
    data2 = Data()

    class CustomWidget(QtWidgets.QWidget):
        def __init__(self, parent=None):

            super(CustomWidget, self).__init__(parent=parent)

            self.layout = QtWidgets.QVBoxLayout()
            self.setLayout(self.layout)

            self.combotext_planet = QtWidgets.QComboBox(objectName='combotext_planet')
            self.layout.addWidget(self.combotext_planet)
            self.combotext_planet.addItem('earth')
            self.combotext_planet.addItem('mars')
            self.combotext_planet.addItem('jupiter')

            self.combodata_dataset = QtWidgets.QComboBox(objectName='combodata_dataset')
            self.layout.addWidget(self.combodata_dataset)
            self.combodata_dataset.addItem('data1', UserDataWrapper(data1))
            self.combodata_dataset.addItem('data2', UserDataWrapper(data2))

            self.text_name = QtWidgets.QLineEdit(objectName='text_name')
            self.layout.addWidget(self.text_name)

            self.valuetext_age = QtWidgets.QLineEdit(objectName='valuetext_age')
            self.layout.addWidget(self.valuetext_age)

            self.value_height = QtWidgets.QSlider(objectName='value_height')
            self.value_height.setMinimum(0)
            self.value_height.setMaximum(10)
            self.layout.addWidget(self.value_height)

            self.bool_log = QtWidgets.QToolButton(objectName='bool_log')
            self.bool_log.setCheckable(True)
            self.layout.addWidget(self.bool_log)

    class Person(object):
        planet = CallbackProperty()
        dataset = CallbackProperty()
        name = CallbackProperty()
        age = CallbackProperty()
        height = CallbackProperty()
        log = CallbackProperty()

    widget = CustomWidget()

    person = Person()

    connect_kwargs = {'height': {'value_range': (0, 100)},
                      'age': {'fmt': '{:.2f}'}}

    c1 = autoconnect_callbacks_to_qt(person, widget, connect_kwargs=connect_kwargs)  # noqa

    # Check that modifying things in the Qt widget updates the callback properties

    widget.combotext_planet.setCurrentIndex(2)
    assert person.planet == 'jupiter'

    widget.combodata_dataset.setCurrentIndex(1)
    assert person.dataset is data2

    widget.text_name.setText('Lovelace')
    widget.text_name.editingFinished.emit()
    assert person.name == 'Lovelace'

    widget.valuetext_age.setText('76')
    widget.valuetext_age.editingFinished.emit()
    assert person.age == 76

    widget.value_height.setValue(7)
    assert person.height == 70

    widget.bool_log.setChecked(True)
    assert person.log

    # Check that modifying the callback properties updates the Qt widget

    person.planet = 'mars'
    assert widget.combotext_planet.currentIndex() == 1

    person.dataset = data1
    assert widget.combodata_dataset.currentIndex() == 0

    person.name = 'Curie'
    assert widget.text_name.text() == 'Curie'

    person.age = 66.3
    assert widget.valuetext_age.text() == '66.30'

    person.height = 54
    assert widget.value_height.value() == 5

    person.log = False
    assert not widget.bool_log.isChecked()


def test_autoconnect_with_empty_qt_item():

    # The following test just makes sure that if a widget without children
    # is ever passed to autoconnect_callbacks_to_qt, things don't crash

    widget = QtGui.QPalette()

    class Person(object):
        name = CallbackProperty()

    person = Person()

    autoconnect_callbacks_to_qt(person, widget)



from pkg_resources import get_distribution, DistributionNotFound

__all__ = ['__version__']

try:
    __version__ = get_distribution('echo').version
except DistributionNotFound:
    __version__ = 'undefined'
